#include "info.h"

#include <errno.h>
#include <stdio.h>
#include "log_utils.h"

int get_volume_info(storage_object_t* region, extended_info_array_t** info_array) {
    LOG_ENTRY();
    LOG_DEBUG("getting basic info for region %s\n", region->name);

    const size_t info_len = 3;
    extended_info_array_t* info =
        EngFncs->engine_alloc(sizeof(extended_info_array_t) + sizeof(extended_info_t) * info_len);
    if (!info)
        return LOG_EXIT(ENOMEM);

    info->count = info_len;
    *info_array = info;

    size_t num = 0;

    /* Region Name */
    info->info[num].name = EngFncs->engine_strdup("Name");
    info->info[num].title = EngFncs->engine_strdup(_("Region Name"));
    info->info[num].desc = EngFncs->engine_strdup(_("Name of this IMSM region (volume)"));
    info->info[num].type = EVMS_Type_String;
    info->info[num].value.s = EngFncs->engine_strdup(region->name);
    ++num;

    /* Container Name */
    info->info[num].name = EngFncs->engine_strdup("Container");
    info->info[num].title = EngFncs->engine_strdup(_("Container Name"));
    info->info[num].desc = EngFncs->engine_strdup(_("Name of the LVM2 container (VG) that produces this region."));
    info->info[num].type = EVMS_Type_String;
    info->info[num].value.s = EngFncs->engine_strdup(region->producing_container->name);
    ++num;

    /* Region Size */
    info->info[num].name = EngFncs->engine_strdup("Size");
    info->info[num].title = EngFncs->engine_strdup(_("Region Size"));
    info->info[num].desc = EngFncs->engine_strdup(_("Space allocated for this region."));
    info->info[num].type = EVMS_Type_Unsigned_Int64;
    info->info[num].unit = EVMS_Unit_Sectors;
    info->info[num].value.ui64 = region->size;
    ++num;

    LOG_EXIT_INT(0);
    return 0;
}

static const char* version_to_str(evms_version_t* version) {
    static char storage[64];
    snprintf(storage, sizeof(storage) - 1, "%d.%d.%d", version->major, version->minor, version->patchlevel);
    return storage;
}

int get_plugin_info(extended_info_array_t** info_array) {
    LOG_ENTRY();

    /* Get memory for the info array. */
    const size_t info_len = 7;
    extended_info_array_t* info =
        EngFncs->engine_alloc(sizeof(extended_info_array_t) + sizeof(extended_info_t) * info_len);
    if (!info)
        return LOG_EXIT(ENOMEM);

    info->count = info_len;
    *info_array = info;

    size_t num = 0;

    /* Short Name. */
    info->info[num].name = EngFncs->engine_strdup("ShortName");
    info->info[num].title = EngFncs->engine_strdup(_("Short Name"));
    info->info[num].desc = EngFncs->engine_strdup(_("A short name given to this plug-in"));
    info->info[num].type = EVMS_Type_String;
    info->info[num].value.s = EngFncs->engine_strdup(my_plugin_record->short_name);
    ++num;

    /* Long Name. */
    info->info[num].name = EngFncs->engine_strdup("LongName");
    info->info[num].title = EngFncs->engine_strdup(_("Long Name"));
    info->info[num].desc = EngFncs->engine_strdup(_("A longer, more descriptive name for this plug-in"));
    info->info[num].type = EVMS_Type_String;
    info->info[num].value.s = EngFncs->engine_strdup(my_plugin_record->long_name);
    ++num;

    /* Plugin Type. */
    info->info[num].name = EngFncs->engine_strdup("Type");
    info->info[num].title = EngFncs->engine_strdup(_("Plugin Type"));
    info->info[num].desc = EngFncs->engine_strdup(
        _("There are various types of plug-ins, each responsible for some kind of storage object or logical volume."));
    info->info[num].type = EVMS_Type_String;
    info->info[num].value.s = EngFncs->engine_strdup(_("Region Manager"));
    ++num;

    /* Plugin Version. */
    info->info[num].name = EngFncs->engine_strdup("Version");
    info->info[num].title = EngFncs->engine_strdup(_("Plugin Version"));
    info->info[num].desc = EngFncs->engine_strdup(_("This is the version number of the plugin."));
    info->info[num].type = EVMS_Type_String;
    info->info[num].value.s = EngFncs->engine_strdup(version_to_str(&my_plugin_record->version));
    ++num;

    /* Required Engine Services Version. */
    info->info[num].name = EngFncs->engine_strdup("Required_Engine_Version");
    info->info[num].title = EngFncs->engine_strdup(_("Required Engine Services Version"));
    info->info[num].desc = EngFncs->engine_strdup(_("This is the version of the Engine services "
                                                  "that this plug-in requires. It will not "
                                                  "run on older versions of the Engine services."));
    info->info[num].type = EVMS_Type_String;
    info->info[num].value.s = EngFncs->engine_strdup(version_to_str(&my_plugin_record->required_engine_api_version));
    ++num;

    /* Required Plug-in API Version. */
    info->info[num].name = EngFncs->engine_strdup("Required_Plugin_Version");
    info->info[num].title = EngFncs->engine_strdup(_("Required Plug-in API Version"));
    info->info[num].desc = EngFncs->engine_strdup(_("This is the version of the Engine plug-in API "
                                                  "that this plug-in requires. It will not run "
                                                  "on older versions of the Engine plug-in API."));
    info->info[num].type = EVMS_Type_String;
    info->info[num].value.s = EngFncs->engine_strdup(version_to_str(&my_plugin_record->required_plugin_api_version.plugin));
    ++num;

    /* Required Container API Version. */
    info->info[num].name = EngFncs->engine_strdup("Required_Container_Version");
    info->info[num].title = EngFncs->engine_strdup(_("Required Container API Version"));
    info->info[num].desc = EngFncs->engine_strdup(_("This is the version of the Engine container "
                                                  "API that this plug-in requires. It will not "
                                                  "run on older versions of the Engine container API."));
    info->info[num].type = EVMS_Type_String;
    info->info[num].value.s = EngFncs->engine_strdup(version_to_str(&my_plugin_record->required_container_api_version));
    ++num;

    return LOG_EXIT(0);
}
